from contextlib import contextmanager

from . import operators as op
from .executors import Breakpoint, Displayer
from .utils import lax_function


def _opmethod(name, operator):
    def f(self, *args, **kwargs):
        return self.pipe(operator(*args, **kwargs))

    f.__name__ = name
    return f


def prox(obs):
    assert not isinstance(obs, ObservableProxy)
    return ObservableProxy(obs)


class ObservableProxy:
    def __init__(self, obs):
        self.obs = obs

    def pipe(self, *args, **kwargs):
        return prox(self.obs.pipe(*args, **kwargs))

    def subscribe(self, *args, **kwargs):
        return self.obs.subscribe(*args, **kwargs)

    def subscribe_(self, *args, **kwargs):
        return self.obs.subscribe_(*args, **kwargs)

    def __rshift__(self, subscription):
        if isinstance(subscription, list):
            subscription = subscription.append
        elif isinstance(subscription, set):
            subscription = subscription.add
        return self.subscribe(subscription)

    def __getitem__(self, item):
        if not isinstance(item, tuple):
            item = (item,)
        strict = not any("?" in x for x in item)
        item = [x.lstrip("?") for x in item]
        return self.getitem(*item, strict=strict)

    all = _opmethod("all", op.all)
    amb = _opmethod("amb", op.amb)
    as_observable = _opmethod("as_observable", op.as_observable)
    average = _opmethod("average", op.average)
    buffer = _opmethod("buffer", op.buffer)
    buffer_toggle = _opmethod("buffer_toggle", op.buffer_toggle)
    buffer_when = _opmethod("buffer_when", op.buffer_when)
    buffer_with_count = _opmethod("buffer_with_count", op.buffer_with_count)
    buffer_with_time = _opmethod("buffer_with_time", op.buffer_with_time)
    buffer_with_time_or_count = _opmethod(
        "buffer_with_time_or_count", op.buffer_with_time_or_count
    )
    cast = _opmethod("cast", op.cast)
    catch = _opmethod("catch", op.catch)
    combine_latest = _opmethod("combine_latest", op.combine_latest)
    concat = _opmethod("concat", op.concat)
    contains = _opmethod("contains", op.contains)
    count = _opmethod("count", op.count)
    datetime = _opmethod("datetime", op.datetime)
    debounce = _opmethod("debounce", op.debounce)
    default_if_empty = _opmethod("default_if_empty", op.default_if_empty)
    delay = _opmethod("delay", op.delay)
    delay_subscription = _opmethod("delay_subscription", op.delay_subscription)
    delay_with_mapper = _opmethod("delay_with_mapper", op.delay_with_mapper)
    dematerialize = _opmethod("dematerialize", op.dematerialize)
    distinct = _opmethod("distinct", op.distinct)
    distinct_until_changed = _opmethod(
        "distinct_until_changed", op.distinct_until_changed
    )
    do = _opmethod("do", op.do)
    do_action = _opmethod("do_action", op.do_action)
    do_while = _opmethod("do_while", op.do_while)
    element_at = _opmethod("element_at", op.element_at)
    element_at_or_default = _opmethod("element_at_or_default", op.element_at_or_default)
    exclusive = _opmethod("exclusive", op.exclusive)
    expand = _opmethod("expand", op.expand)
    filter = _opmethod("filter", op.filter)
    filter_indexed = _opmethod("filter_indexed", op.filter_indexed)
    finally_action = _opmethod("finally_action", op.finally_action)
    find = _opmethod("find", op.find)
    find_index = _opmethod("find_index", op.find_index)
    first = _opmethod("first", op.first)
    first_or_default = _opmethod("first_or_default", op.first_or_default)
    flat_map = _opmethod("flat_map", op.flat_map)
    flat_map_indexed = _opmethod("flat_map_indexed", op.flat_map_indexed)
    flat_map_latest = _opmethod("flat_map_latest", op.flat_map_latest)
    fork_join = _opmethod("fork_join", op.fork_join)
    group_by = _opmethod("group_by", op.group_by)
    group_by_until = _opmethod("group_by_until", op.group_by_until)
    group_join = _opmethod("group_join", op.group_join)
    ignore_elements = _opmethod("ignore_elements", op.ignore_elements)
    is_empty = _opmethod("is_empty", op.is_empty)
    join = _opmethod("join", op.join)
    last = _opmethod("last", op.last)
    last_or_default = _opmethod("last_or_default", op.last_or_default)
    map = _opmethod("map", op.map)
    map_indexed = _opmethod("map_indexed", op.map_indexed)
    materialize = _opmethod("materialize", op.materialize)
    max = _opmethod("max", op.max)
    max_by = _opmethod("max_by", op.max_by)
    merge = _opmethod("merge", op.merge)
    merge_all = _opmethod("merge_all", op.merge_all)
    min = _opmethod("min", op.min)
    min_by = _opmethod("min_by", op.min_by)
    multicast = _opmethod("multicast", op.multicast)
    observe_on = _opmethod("observe_on", op.observe_on)
    on_error_resume_next = _opmethod("on_error_resume_next", op.on_error_resume_next)
    overload = _opmethod("overload", op.overload)
    pairwise = _opmethod("pairwise", op.pairwise)
    partition = _opmethod("partition", op.partition)
    partition_indexed = _opmethod("partition_indexed", op.partition_indexed)
    pluck = _opmethod("pluck", op.pluck)
    pluck_attr = _opmethod("pluck_attr", op.pluck_attr)
    publish = _opmethod("publish", op.publish)
    publish_value = _opmethod("publish_value", op.publish_value)
    reduce = _opmethod("reduce", op.reduce)
    ref_count = _opmethod("ref_count", op.ref_count)
    repeat = _opmethod("repeat", op.repeat)
    replay = _opmethod("replay", op.replay)
    retry = _opmethod("retry", op.retry)
    sample = _opmethod("sample", op.sample)
    scan = _opmethod("scan", op.scan)
    sequence_equal = _opmethod("sequence_equal", op.sequence_equal)
    share = _opmethod("share", op.share)
    single = _opmethod("single", op.single)
    single_or_default = _opmethod("single_or_default", op.single_or_default)
    single_or_default_async = _opmethod(
        "single_or_default_async", op.single_or_default_async
    )
    skip = _opmethod("skip", op.skip)
    skip_last = _opmethod("skip_last", op.skip_last)
    skip_last_with_time = _opmethod("skip_last_with_time", op.skip_last_with_time)
    skip_until = _opmethod("skip_until", op.skip_until)
    skip_until_with_time = _opmethod("skip_until_with_time", op.skip_until_with_time)
    skip_while = _opmethod("skip_while", op.skip_while)
    skip_while_indexed = _opmethod("skip_while_indexed", op.skip_while_indexed)
    skip_with_time = _opmethod("skip_with_time", op.skip_with_time)
    slice = _opmethod("slice", op.slice)
    some = _opmethod("some", op.some)
    starmap = _opmethod("starmap", op.starmap)
    starmap_indexed = _opmethod("starmap_indexed", op.starmap_indexed)
    start_with = _opmethod("start_with", op.start_with)
    subscribe_on = _opmethod("subscribe_on", op.subscribe_on)
    sum = _opmethod("sum", op.sum)
    switch_latest = _opmethod("switch_latest", op.switch_latest)
    take = _opmethod("take", op.take)
    take_last = _opmethod("take_last", op.take_last)
    take_last_buffer = _opmethod("take_last_buffer", op.take_last_buffer)
    take_last_with_time = _opmethod("take_last_with_time", op.take_last_with_time)
    take_until = _opmethod("take_until", op.take_until)
    take_until_with_time = _opmethod("take_until_with_time", op.take_until_with_time)
    take_while = _opmethod("take_while", op.take_while)
    take_while_indexed = _opmethod("take_while_indexed", op.take_while_indexed)
    take_with_time = _opmethod("take_with_time", op.take_with_time)
    throttle_first = _opmethod("throttle_first", op.throttle_first)
    throttle_with_mapper = _opmethod("throttle_with_mapper", op.throttle_with_mapper)
    throttle_with_timeout = _opmethod("throttle_with_timeout", op.throttle_with_timeout)
    time_interval = _opmethod("time_interval", op.time_interval)
    timedelta = _opmethod("timedelta", op.timedelta)
    timeout = _opmethod("timeout", op.timeout)
    timeout_with_mapper = _opmethod("timeout_with_mapper", op.timeout_with_mapper)
    timestamp = _opmethod("timestamp", op.timestamp)
    to_dict = _opmethod("to_dict", op.to_dict)
    to_future = _opmethod("to_future", op.to_future)
    to_iterable = _opmethod("to_iterable", op.to_iterable)
    to_list = _opmethod("to_list", op.to_list)
    to_marbles = _opmethod("to_marbles", op.to_marbles)
    to_set = _opmethod("to_set", op.to_set)
    typing = _opmethod("typing", op.typing)
    while_do = _opmethod("while_do", op.while_do)
    window = _opmethod("window", op.window)
    window_toggle = _opmethod("window_toggle", op.window_toggle)
    window_when = _opmethod("window_when", op.window_when)
    window_with_count = _opmethod("window_with_count", op.window_with_count)
    window_with_time = _opmethod("window_with_time", op.window_with_time)
    window_with_time_or_count = _opmethod(
        "window_with_time_or_count", op.window_with_time_or_count
    )
    with_latest_from = _opmethod("with_latest_from", op.with_latest_from)
    zip = _opmethod("zip", op.zip)
    zip_with_iterable = _opmethod("zip_with_iterable", op.zip_with_iterable)
    zip_with_list = _opmethod("zip_with_list", op.zip_with_list)

    affix = _opmethod("affix", op.affix)
    as_ = _opmethod("as_", op.as_)
    average_and_variance = _opmethod("average_and_variance", op.average_and_variance)
    collect_between = _opmethod("collect_between", op.collect_between)
    format = _opmethod("format", op.format)
    getitem = _opmethod("getitem", op.getitem)
    keyfilter = _opmethod("keyfilter", op.keyfilter)
    keymap = _opmethod("keymap", op.keymap)
    rekey = _opmethod("rekey", op.rekey)
    roll = _opmethod("roll", op.roll)
    stream_once = _opmethod("stream_once", op.stream_once)
    tag = _opmethod("tag", op.tag)
    throttle = _opmethod("throttle", op.throttle)
    unique = _opmethod("unique", op.unique)
    variance = _opmethod("variance", op.variance)
    where = _opmethod("where", op.where)

    def display(self, *, breakword=False, word=None, **kwargs):
        sub = self.subscribe(Displayer(**kwargs))
        if breakword:  # pragma: no cover
            self.breakword(word=word)
        return sub

    def breakpoint(self):  # pragma: no cover
        return self.subscribe(Breakpoint())

    def breakword(self, **kwargs):  # pragma: no cover
        return self.subscribe(Breakpoint(use_breakword=True, **kwargs))

    def give(self, *keys, **extra):
        from .core import Giver

        giver = Giver(keys=keys, extra=extra)

        if len(keys) == 0:
            gv = giver.produce
        elif len(keys) == 1:
            gv = giver
        else:

            def gv(x):
                return giver(*x)

        return self.subscribe(gv)

    def keysubscribe(self, fn):
        fn = lax_function(fn)
        self.subscribe(lambda data: fn(**data))

    def wrap(self, fn, begin="$begin", end="$end"):
        fn = contextmanager(lax_function(fn))

        managers = {}

        def watch(data):
            if begin in data:
                key = data[begin]
                assert key not in data
                managers[key] = fn(**data)
                managers[key].__enter__()

            if end in data:
                key = data[end]
                managers[key].__exit__(None, None, None)

        return self.subscribe(watch)



from . import operators as op
from .core import (
    Given,
    Giver,
    accumulate,
    give,
    given,
    giver,
    make_give,
    register_special,
    resolve,
)
from .obs import ObservableProxy, prox
from .version import version



import builtins

from rx import operators as rxop
from rx.operators import NotSet

from .utils import keyword_decorator, lax_function

###########
# Reducer #
###########


class _Reducer:
    def __init__(self, reduce, roll):
        self.reduce = reduce
        self.roll = roll


@keyword_decorator
def reducer(func, default_seed=NotSet, postprocess=NotSet):
    name = func.__name__
    if isinstance(func, type):
        constructor = func

    else:

        def constructor():
            return _Reducer(reduce=func, roll=None)

    def _create(*args, scan=False, seed=NotSet, **kwargs):
        reducer = constructor(*args, **kwargs)

        if seed is NotSet:
            seed = default_seed

        if scan is True:
            oper = rxop.scan(reducer.reduce, seed=seed)

        elif scan:
            oper = roll(n=scan, reduce=reducer.roll, seed=seed)

        else:
            oper = rxop.reduce(reducer.reduce, seed=seed)

        if postprocess is not NotSet:
            oper = rxop.pipe(oper, postprocess)

        return oper

    _create.__name__ = name
    return _create


##############################
# Extra operators for giving #
##############################


def affix(**streams):
    """Augment a stream of dicts with extra keys.

    The affixed streams should have the same length as the main one, so when
    affixing a reduction, one should set scan=True, or the roll option.

    Example:
        obs.affix(
            minx=obs["x"].min(scan=True),
            xpy=obs["x", "y"].starmap(lambda x, y: x + y),
        )

    Arguments:
        streams: A mapping from extra keys to add to the dicts to Observables
            that generate the values.
    """

    keys = list(streams.keys())
    values = list(streams.values())

    def merge(elems):
        main, *rest = elems
        return {**main, **dict(zip(keys, rest))}

    return rxop.pipe(rxop.zip(*values), rxop.map(merge))


def as_(key):
    """Make a stream of dictionaries using the given key.

    For example, [1, 2].as_("x") => [{"x": 1}, {"x": 2}]
    """
    return rxop.map(lambda x: {key: x})


@reducer(default_seed=(0, 0), postprocess=rxop.starmap(lambda x, sz: x / sz))
class average:
    def reduce(self, last, add):
        x, sz = last
        return (x + add, sz + 1)

    def roll(self, last, add, drop, last_size, current_size):
        x, _ = last
        if last_size == current_size:
            return (x + add - drop, current_size)
        else:
            return (x + add, current_size)


def _average_and_variance_postprocess(sm, v2, sz):
    avg = sm / sz
    if sz >= 2:
        var = v2 / (sz - 1)
    else:
        var = None
    return (avg, var)


@reducer(
    default_seed=(0, 0, 0),
    postprocess=rxop.starmap(_average_and_variance_postprocess),
)
class average_and_variance:
    def reduce(self, last, add):
        prev_sum, prev_v2, prev_size = last
        new_size = prev_size + 1
        new_sum = prev_sum + add
        if prev_size:
            prev_mean = prev_sum / prev_size
            new_mean = new_sum / new_size
            new_v2 = prev_v2 + (add - prev_mean) * (add - new_mean)
        else:
            new_v2 = prev_v2
        return (new_sum, new_v2, new_size)

    def roll(self, last, add, drop, last_size, current_size):
        if last_size == current_size:
            prev_sum, prev_v2, prev_size = last
            new_sum = prev_sum - drop + add
            prev_mean = prev_sum / prev_size
            new_mean = new_sum / prev_size
            new_v2 = (
                prev_v2
                + (add - prev_mean) * (add - new_mean)
                - (drop - prev_mean) * (drop - new_mean)
            )
            return (new_sum, new_v2, prev_size)
        else:
            return self.reduce(last, add)


def collect_between(start, end, common=None):
    """Collect all data between the start and end keys.

    Example:
        with given() as gv:
            gv.collect_between("A", "Z") >> (results := [])
            give(A=1)
            give(B=2)
            give(C=3, D=4, A=5)
            give(Z=6)
            assert results == [{"A": 5, "B": 2, "C": 3, "D": 4, "Z": 6}]

    Arguments:
        start: The key that marks the beginning of the accumulation.
        end: The key that marks the end of the accumulation.
        common: A key that must be present in all data and must have
            the same value in the whole group.
    """
    import rx

    def aggro(source):
        def subscribe(obs, scheduler=None):
            dicts = {}

            def on_next(value):
                if isinstance(value, dict):
                    if common is not None:
                        if common in value:
                            key = value[common]
                        else:
                            return
                    else:
                        key = None
                    if start in value:
                        dicts[key] = dict(value)
                    elif end in value:
                        current = dicts.setdefault(key, {})
                        current.update(value)
                        obs.on_next(current)
                        del dicts[key]
                    else:
                        current = dicts.setdefault(key, {})
                        current.update(value)

            return source.subscribe(on_next, obs.on_error, obs.on_completed, scheduler)

        return rx.create(subscribe)

    return aggro


@reducer(default_seed=0)
class count:
    def __init__(self, predicate=None):
        self.predicate = predicate

    def reduce(self, last, new):
        if self.predicate is None or self.predicate(new):
            return last + 1
        else:
            return last

    def roll(self, last, new, drop, last_size, current_size):
        if last_size == current_size:
            if self.predicate is None:
                return last
            else:
                plus = 1 if self.predicate(new) else 0
                minus = 1 if self.predicate(drop) else 0
                return last + plus - minus
        else:
            return self.reduce(last, new)


def format(string):
    """Format an object using a format string.

    Arguments:
        string: The format string.
    """

    def _fmt(x):
        if isinstance(x, dict):
            return string.format(**x)
        elif isinstance(x, (list, tuple)):
            return string.format(*x)
        else:
            return string.format(x)

    return rxop.map(_fmt)


def getitem(*names, strict=False):
    """Extract a key from a dictionary.

    Arguments:
        name: Name of the key to index with.
        strict: If true, every element in the stream is required to
            contains this key.
    """
    import operator

    if len(names) == 1:
        (name,) = names
        if strict:
            return rxop.map(operator.itemgetter(name))
        else:
            return rxop.pipe(
                rxop.filter(lambda arg: name in arg),
                rxop.map(operator.itemgetter(name)),
            )
    else:
        if strict:
            return rxop.map(lambda arg: tuple(arg[name] for name in names))
        else:
            return rxop.pipe(
                rxop.filter(lambda arg: builtins.all(name in arg for name in names)),
                rxop.map(lambda arg: tuple(arg[name] for name in names)),
            )


def keyfilter(fn=None):
    fn = lax_function(fn)
    return rxop.filter(lambda kwargs: fn(**kwargs))


def keymap(_fn=None, **_kwargs):
    """Map a dict, passing keyword arguments.

    keymap either takes a positional function argument or keyword arguments
    serving to build a new dict.

    Arguments:
        _fn: A function that will be called for each element, passing the
            element using **kwargs.

            Note: If the dict has elements that are not in the function's
            arguments list and the function does not have a **kwargs
            argument, these elements will be dropped and no error will
            occur.
        _kwargs: Alternatively, build a new dict with each key associated to
            a function with the same interface as fn.
    """
    if _fn and _kwargs or not _fn and not _kwargs:
        raise TypeError(
            "keymap either takes one argument or keyword arguments but not both"
        )

    elif _fn:
        _fn = lax_function(_fn)
        return rxop.map(lambda kwargs: _fn(**kwargs))

    else:
        fns = {k: lax_function(fn) for k, fn in _kwargs.items()}
        return rxop.map(lambda kwargs: {k: fn(**kwargs) for k, fn in fns.items()})


@reducer
def min(last, new):
    return builtins.min(last, new)


@reducer
def max(last, new):
    return builtins.max(last, new)


def rekey(*keep, **remap):
    """Keep certain dict keys and remap others.

    Arguments:
        keep: Keys that must be kept
        remap: Keys that must be renamed
    """
    remap = {**{k: k for k in keep}, **remap}

    def _rekey(data):
        return {k2: data[k1] for k1, k2 in remap.items()}

    return rxop.map(_rekey)


def roll(n, reduce=None, seed=NotSet):  # noqa: F811
    """Group the last n elements, giving a sequence of overlapping sequences.

    This can be used to compute a rolling average of the 100 last element:
        op.roll(100, lambda xs: sum(xs) / len(xs))

    Arguments:
        n: The number of elements to group together.
        reduce: A function to reduce the group.

            It should take four arguments:
                last: The last result.
                add: The element that was just added. It is the last element
                    in the elements list.
                drop: The element that was dropped to make room for the
                    added one. It is *not* in the elements argument.
                    If the list of elements is not yet of size n, there is
                    no need to drop anything and drop is None.
                last_size: The window size on the last invocation.
                current_size: The window size on this invocation.

            Defaults to returning the deque of elements directly. The same
            reference is returned each time in order to save memory, so it
            should be processed immediately.
        seed: The first element of the reduction.
    """

    from collections import deque

    q = deque(maxlen=n)

    if reduce is not None:

        def queue(current, x):
            drop = q[0] if len(q) == n else NotSet
            last_size = len(q)
            q.append(x)
            current_size = len(q)
            return reduce(
                current,
                x,
                drop=drop,
                last_size=last_size,
                current_size=current_size,
            )

        scan_command = rxop.scan(queue, seed)

    else:

        def queue(q, x):
            q.append(x)
            return q

        scan_command = rxop.scan(queue, q)

    return rxop.pipe(scan_command, stream_once())


def stream_once():
    """Make sure that upstream operators only run once.

    Use this if upstream operators have side effects, otherwise each
    downstream subscription will re-run the effects.
    """

    import rx

    def go(source):
        observers = []

        def on_next(value):
            for obv in observers:
                obv.on_next(value)

        def on_error(value):  # pragma: no cover
            for obv in observers:
                obv.on_error(value)

        def on_completed():
            for obv in observers:
                obv.on_completed()

        def subscribe(obv, scheduler):
            observers.append(obv)
            return dispo

        dispo = source.subscribe_(on_next, on_error, on_completed)
        return rx.Observable(subscribe)

    return go


@reducer
def sum(last, new):
    return last + new


def tag(group="", field="$word", group_field="$group"):
    """Tag each dict or object with a unique word.

    If the item is a dict, do `item[field] = <new_word>`, otherwise
    attempt to do `setattr(item, field, <new_word>)`.

    These tags are displayed specially by the `display` method and they
    can be used to determine breakpoints with the `breakword` method.

    Arguments:
        group: An arbitrary group name that corresponds to an independent
            sequence of words. It determines the color in display.
        field: The field name in which to put the word
            (default: `$word`).
        group_field: The field name in which to put the group
            (default: `$group`).
    """
    try:
        import breakword as bw
    except ImportError:  # pragma: no cover
        raise ImportError(
            "Package `breakword` must be installed to use the tag() operator"
        )

    grp = bw.groups[group]

    def tag_data(data):
        word = grp.gen()
        if isinstance(data, dict):
            data = {field: word, **data}
            if group:
                data[group_field] = group
        else:
            if group:
                setattr(data, group_field, group)
                setattr(data, "$group", group)
            setattr(data, field, word)
            setattr(data, "$word", word)
        return data

    return rxop.pipe(rxop.map(tag_data), stream_once())


def unique():
    """Collect unique elements.

    Be aware that this keeps a set of all the elements seen so far,
    so it may prevent them from being reclaimed by garbage collection
    and can be expensive in memory.
    """
    import rx

    def oper(source):
        def subscribe(obv, scheduler=None):
            elements = set()

            def on_next(value):
                if isinstance(value, dict):
                    key = tuple(value.items())
                else:
                    key = value

                if key not in elements:
                    elements.add(key)
                    obv.on_next(value)

            return source.subscribe(on_next, obv.on_error, obv.on_completed, scheduler)

        return rx.create(subscribe)

    return oper


def variance(*args, **kwargs):
    return rxop.pipe(
        average_and_variance(*args, **kwargs), rxop.starmap(lambda avg, var: var)
    )


def where(*keys, **conditions):
    """Filter entries with the given keys meeting the given conditions.

    Example:
        where("x", "!y", z=True, w=lambda x: x > 0)

    Arguments:
        keys: Keys that must be present in the dictionary or, if a key starts
            with "!", it must *not* be present.
        conditions: Maps a key to the value it must be associated to in the
            dictionary, or to a predicate function on the value.
    """
    from types import FunctionType

    conditions = {
        k: cond if isinstance(cond, FunctionType) else lambda x: cond == x
        for k, cond in conditions.items()
    }
    excluded = [k.lstrip("!") for k in keys if k.startswith("!")]
    keys = [k for k in keys if not k.startswith("!")]
    keys = (*keys, *conditions.keys())

    def filt(x):
        return (
            isinstance(x, dict)
            and builtins.all(k in x for k in keys)
            and not builtins.any(k in x for k in excluded)
            and builtins.all(cond(x[k]) for k, cond in conditions.items())
        )

    return rxop.filter(filt)



import functools
import types


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=None, **kwargs):
        if fn is None:

            @functools.wraps(deco)
            def newer_deco(fn):
                return deco(fn, **kwargs)

            return newer_deco
        else:
            return deco(fn, **kwargs)

    return new_deco


def lax_function(fn):
    if isinstance(fn, types.FunctionType):
        KWVAR_FLAG = 8
        co = fn.__code__
        if not co.co_flags & KWVAR_FLAG:
            if hasattr(co, "replace"):
                newco = co.replace(
                    co_flags=co.co_flags | KWVAR_FLAG,
                    # Add a dummy keyword argument with an illegal name
                    co_varnames=(*co.co_varnames, "#"),
                )
            else:  # pragma: no cover
                newco = types.CodeType(
                    co.co_argcount,
                    co.co_kwonlyargcount,
                    co.co_nlocals,
                    co.co_stacksize,
                    co.co_flags | KWVAR_FLAG,
                    co.co_code,
                    co.co_consts,
                    co.co_names,
                    (*co.co_varnames, "#"),
                    co.co_filename,
                    co.co_name,
                    co.co_firstlineno,
                    co.co_lnotab,
                    co.co_freevars,
                    co.co_cellvars,
                )

            newfn = types.FunctionType(
                name=fn.__name__,
                code=newco,
                globals=fn.__globals__,
                closure=fn.__closure__,
            )
            newfn.__defaults__ = fn.__defaults__
            newfn.__kwdefaults__ = fn.__kwdefaults__
            return newfn

    return fn



import ast
import sys
import time
from collections import namedtuple
from contextlib import contextmanager
from contextvars import ContextVar
from itertools import count
from types import SimpleNamespace

import rx
from varname import ImproperUseError, VarnameRetrievingError, argname, varname
from varname.utils import get_node

from . import operators as op
from .obs import ObservableProxy

ABSENT = object()

global_context = ContextVar("global_context", default=())
global_inherited = ContextVar("global_inherited", default={})

_block_classes = {
    ast.If: ("body", "orelse"),
    ast.For: ("body", "orelse"),
    ast.While: ("body", "orelse"),
    ast.FunctionDef: ("body",),
    ast.AsyncFunctionDef: ("body",),
    ast.With: ("body",),
    ast.AsyncWith: ("body",),
    ast.AsyncFor: ("body", "orelse"),
}


_improper_nullary_give_error = (
    "give() with no arguments must immediately follow an assignment"
)


special_keys = {}


global_count = count(0)


def register_special(key):
    def deco(func):
        special_keys[key] = func
        return func

    return deco


@register_special("$time")
def _special_time():
    return time.time()


@register_special("$frame")
def _special_frame():
    return sys._getframe(3)


LinePosition = namedtuple("LinePosition", ["name", "filename", "lineno"])


@register_special("$line")
def _special_line():
    fr = sys._getframe(3)
    co = fr.f_code
    return LinePosition(co.co_name, co.co_filename, fr.f_lineno)


def _find_above(frame):
    node = get_node(frame + 1)
    if node is None:
        raise VarnameRetrievingError(
            "Cannot retrieve the node where the function is called"
        )

    while node.parent is not None:
        parent = node.parent
        fields = _block_classes.get(type(parent), None)
        if fields is None:
            node = parent
            continue
        else:
            for field in fields:
                f = getattr(parent, field)
                if node in f:
                    idx = f.index(node)
                    if idx == 0:
                        raise ImproperUseError(_improper_nullary_give_error)

                    assignment = f[idx - 1]

                    if isinstance(assignment, ast.Assign):
                        target = assignment.targets[-1]
                        names = [target.id]
                    elif isinstance(assignment, (ast.AugAssign, ast.AnnAssign)):
                        names = [assignment.target.id]
                    else:
                        raise ImproperUseError(_improper_nullary_give_error)

                    fr = sys._getframe(frame)
                    rval = {}

                    for name in names:
                        if name in fr.f_locals:
                            rval[name] = fr.f_locals[name]
                        elif name in fr.f_globals:
                            rval[name] = fr.f_globals[name]
                        else:  # pragma: no cover
                            # I am not sure how to trigger this
                            raise Exception("Could not resolve value")
                    return rval

            else:  # pragma: no cover
                # I am not sure how to trigger this
                raise Exception("Could not find node position")

    # I am not sure how to trigger this
    raise Exception("Could not find node")  # pragma: no cover


def resolve(frame, func, args):
    """Return a {variable_name: value} dictionary depending on usage.

    * (len(args) == 0) => Use the variable assigned in the line before the call.
    * (len(args) == 1) => Use the variable the call is assigned to.
    * (len(args) >= 1) => Use the variables passed as arguments to the call.
    """
    nargs = len(args)

    if nargs == 0:
        return _find_above(frame=frame + 2)

    if nargs == 1:
        try:
            assigned_to = varname(frame=frame + 1, strict=True, raise_exc=False)
        except ImproperUseError:
            assigned_to = None
        if assigned_to is not None:
            return {assigned_to: args[0]}

    argnames = argname("args", func=func, frame=frame + 1, vars_only=False)
    if argnames is None:  # pragma: no cover
        # I am not sure how to trigger this
        raise Exception("Could not resolve arg names")

    return {name: value for name, value in zip(argnames, args)}


class Giver:
    def __init__(
        self,
        *,
        keys=None,
        special=[],
        extra={},
        context=global_context,
        inherited=global_inherited,
    ):
        self.keys = keys
        self.special = special
        self.extra = extra
        self.context = context
        self.inherited = inherited

    @property
    def line(self):
        return Giver(
            keys=self.keys,
            special=(*self.special, "$line"),
            extra=self.extra,
            context=self.context,
            inherited=self.inherited,
        )

    @property
    def time(self):
        return Giver(
            keys=self.keys,
            special=(*self.special, "$time"),
            extra=self.extra,
            context=self.context,
            inherited=self.inherited,
        )

    @contextmanager
    def inherit(self, **keys):
        inh = self.inherited.get()
        token = self.inherited.set({**inh, **keys})
        try:
            yield
        finally:
            self.inherited.reset(token)

    @contextmanager
    def wrap(self, **keys):
        with self.wrap_no_inherit(**keys):
            with self.inherit(**keys):
                yield

    @contextmanager
    def wrap_no_inherit(self, **keys):
        num = next(global_count)
        self.produce({"$begin": num, **keys})
        try:
            yield
        finally:
            self.produce({"$end": num, **keys})

    def produce(self, values):
        for special in self.special:
            values[special] = special_keys[special]()

        if self.extra:
            values = {**self.extra, **values}

        inh = self.inherited.get()
        if inh is not None:
            values = {**inh, **values}

        for handler in self.context.get():
            handler(values)

    def __call__(self, *args, **values):
        h = self.context.get()
        if h:
            if self.keys:
                if len(args) != len(self.keys):
                    raise ImproperUseError(
                        f"Giver for {self.keys} must have {len(self.keys)} positional argument(s)."
                    )
                keyed = dict(zip(self.keys, args))
                values = {**keyed, **values}
            elif args:
                values = {**resolve(1, self, args), **values}
            elif not values:
                values = resolve(1, self, ())

            self.produce(values)

        if len(args) == 1:
            return args[0]
        else:
            return None


def giver(*keys, **extra):
    normal = [k for k in keys if not k.startswith("$")]
    special = [k for k in keys if k.startswith("$")]
    return Giver(keys=normal, special=special, extra=extra)


class Given:
    def __init__(self, key=None, context=global_context):
        self.key = key
        self.token = self.observers = None
        self.context = context

    def __enter__(self):
        self.observers = []

        def make(observer, scheduler):
            self.observers.append(observer)

        src = rx.create(make)

        def handler(values):
            for obs in self.observers:
                obs.on_next(values)

        h = self.context.get()
        self.token = self.context.set((*h, handler))

        if isinstance(self.key, str):
            src = src.pipe(op.getitem(self.key))

        return ObservableProxy(src)

    def __exit__(self, exc_type=None, exc=None, tb=None):
        for obs in self.observers:
            obs.on_completed()
        self.context.reset(self.token)
        self.token = self.observers = None


def make_give(context=None):
    context = context or ContextVar("context", default=())
    give = Giver(context=context)

    def given(*args, **kwargs):
        return Given(*args, **kwargs, context=context)

    @contextmanager
    def accumulate(key=None):
        results = []
        with given(key) as gv:
            gv.subscribe(results.append)
            yield results

    return SimpleNamespace(
        context=context,
        give=give,
        given=given,
        accumulate=accumulate,
    )


_global_given = make_give(context=global_context)

give = _global_given.give
given = _global_given.given
accumulate = _global_given.accumulate
